//
// Copyright 2014 Perforce Software Inc.
//

using Perforce.View;
using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using System.Windows;
using System.Windows.Controls;
using System.Runtime.Remoting.Metadata.W3cXsd2001;
using Perforce.P4;
using Perforce.ViewModel;
using Perforce.Model;

namespace Perforce.Helper {
    public class ListingHelper {

        // method to get listings from Perforce
        public static IList<ListingItem> GetDepotListing(ColumnDisplay column, string basePath) {
            if (string.IsNullOrEmpty(basePath)) {
                basePath = "//";
            }
            //UIHelper.SetBusyState();
            var list = new List<ListingItem>();
            //var helper = Utility.GetPerforceHelper();
            var helper = Utility.GetPerforceHelper(check:true);
            if (basePath.Equals("//")) {
                // get list of available depots (local only -- no streams)
                var depotList = helper.ListDepots(true);
                foreach (var depot in depotList) {
                    var item = new FolderItem();
                    item.LabelText = depot.Id;
                    item.ColumnContext = column;
                    item.DepotPath = string.Format("//{0}", depot.Id);
                    var paths = new List<string>();
                    paths.Add(item.DepotPath);
                    var mappings = helper.GetClientMappings(paths, pathsAreDirectories: true);
                    if (mappings.ContainsKey(item.DepotPath)) {
                        var mapping = mappings[item.DepotPath];
                        item.ClientPath = mapping.Remove(mapping.Length-1);
                    }
                    item.IsMapped = helper.IsDirectoryMapped(item.DepotPath);
                    item.IsPathInClientView = helper.IsPathInClientView(item.DepotPath);
                    list.Add(item);
                }
            } else {
                var dirList = helper.ListDirectories(basePath);
                var mappings = helper.GetClientMappings(dirList, true);
                foreach (var dir in dirList) {
                    var dirName = dir.Replace(basePath, string.Empty);
                    if (dirName.StartsWith("/")) {
                        dirName = dirName.Remove(0, 1);
                    }
                    var item = new FolderItem();
                    item.LabelText = dirName;
                    item.ColumnContext = column;
                    item.DepotPath = dir;
                    item.IsMapped = mappings.ContainsKey(dir);
                    if (item.IsMapped) {
                        var mapping = mappings[dir];
                        item.ClientPath = mapping.Remove(mapping.Length-1);
                    }
                    item.IsPathInClientView = helper.IsPathInClientView(item.DepotPath);
                    list.Add(item);
                }
                var metadataList = helper.ListFiles(depotPath: basePath + "/*", showDeleted: false);
                if (metadataList != null) {
                    foreach (var md in metadataList) {
                        var item = new FileItem();
                        item.LabelText = md.GetFileName();
                        item.Metadata = md;
                        item.ColumnContext = column;
                        list.Add(item);
                    }
                }
            }
            return list;
        }

        // method to get listings from the filesystem
        public static IList<ListingItem> GetDirectoryListing(ColumnDisplay column, string dirName) {
            var helper = Utility.GetPerforceHelper();
            if (string.IsNullOrEmpty(dirName)) {
                dirName = helper.CurrentClient.Root;
            }
            //UIHelper.SetBusyState();
            var list = new List<ListingItem>();
            var di = new DirectoryInfo(dirName);
            if (di.Exists) {
                var dirs = di.GetDirectories();
                var paths = new List<string>();
                foreach (var d in dirs) {
                    paths.Add(d.FullName);
                }
                var mappings = helper.GetDepotMappings(paths, true);
                foreach (var d in dirs) {
                    var item = new FolderItem();
                    item.LabelText = d.Name;
                    item.ColumnContext = column;
                    item.ClientPath = d.FullName;
                    var normalizedPath = Utility.NormalizeClientPath(d.FullName);
                    item.IsMapped = mappings.ContainsKey(normalizedPath);
                    item.DepotPath = helper.PredictDepotPath(d.FullName);
                    item.IsPathInClientView = helper.IsPathInClientView(item.DepotPath);
                    list.Add(item);
                }
                var files = di.GetFiles();
                var fileList = new List<string>();
                foreach (var f in files) {
                    fileList.Add(f.FullName);
                }
                var results = helper.GetFileMetaData(fileList);
                var metadataMap = new Dictionary<string, FileMetaData>(StringComparer.CurrentCultureIgnoreCase);
                if (results != null) {
                    foreach (var md in results) {
                        metadataMap.Add(md.ClientPath.Path, md);
                    }
                }
                foreach (var f in files) {
                    var item = new FileItem();
                    item.LabelText = f.Name;
                    item.ColumnContext = column;
                    if (metadataMap.ContainsKey(f.FullName)) {
                        item.Metadata = metadataMap[f.FullName];
                    } else {
                        // fake out the file metadata
                        var metadata = new FileMetaData();
                        metadata.IsMapped = false;
                        metadata.DepotPath = new DepotPath(Constants.DUMMY_DEPOT_PATH);
                        metadata.ClientPath = new ClientPath(f.FullName);
                        item.Metadata = metadata;
                    }
                    list.Add(item);
                }
            }
            return list;
        }

        
        public static IList<ListingItem> GetSearchResults(ColumnDisplay column, string[] tagFilter = null) {
            var list = new List<ListingItem>();
            // get the search view model
            var model = column.Selector.Model as SearchSelectorViewModel;
            if (model.SearchStr != null) {
                // get the search helper
                var searchHelper = Utility.GetSearchHelper();
                if (searchHelper != null) {
                    var results = searchHelper.DoSearch(model.SearchStr, files: model.FileSearch, content: model.ContentSearch, tags: model.TagSearch);
                    if (results.Count > 0) {
                        var paths = new List<string>();
                        foreach (var r in results) {
                            paths.Add(r.depotFile);
                        }
                        List<string> uniquePaths = paths.Distinct<string>().ToList();
                        var helper = Utility.GetPerforceHelper();
                        var metadataList = helper.GetSearchFileMetaData(uniquePaths, mappedOnly: model.MyWorkspaceLocation);
                        if (metadataList != null) {
                            foreach (var md in metadataList) {
                                if (model.RestrictSearchToFolder) {
                                    if (!md.DepotPath.Path.StartsWith(model.DepotPath, StringComparison.InvariantCultureIgnoreCase)) continue;
                                }
                                if (tagFilter != null && tagFilter.Count() > 0) {
                                    // skip if there are no tags
                                    if (!md.Attributes.ContainsKey("tags")) continue;
                                    var hexTags = md.Attributes["tags"] as string;
                                    var tagList = ListingHelper.HexToString(hexTags).Split(',');
                                    var foundAllTagFilters = true;
                                    foreach (var t in tagFilter) {
                                        if (!string.IsNullOrEmpty(t)) {
                                            if (!tagList.Contains(t)) foundAllTagFilters = false;
                                        }
                                    }
                                    if (!foundAllTagFilters) continue;
                                }
                                var item = new FileItem();
                                item.LabelText = md.GetFileName();
                                item.Metadata = md;
                                item.ColumnContext = column;
                                list.Add(item);
                            }
                        }
                    }
                }
            }
            return list;
        }

        public static IList<ListingItem> GetChangelistFiles(ColumnDisplay column) {
            var list = new List<ListingItem>();
            var helper = Utility.GetPerforceHelper();
            if (helper != null && helper.ClientEnabled) {
                var changeFiles = helper.GetChangelistFiles();
                if (changeFiles != null && changeFiles.Count > 0) {
                    foreach (var f in changeFiles) {
                        if (f.Action.Equals(FileAction.Delete)) continue;
                        var item = new FileItem();
                        item.LabelText = f.GetFileName();
                        item.ColumnContext = column;
                        item.Metadata = helper.GetFileMetaData(f.DepotPath.Path);
                        list.Add(item);
                    }
                }
            }
            return list;
        }

        public static IList<ListingItem> GetTrashFiles(ColumnDisplay column) {
            var list = new List<ListingItem>();
            var helper = Utility.GetPerforceHelper();
            if (helper != null && helper.ClientEnabled) {
                //UIHelper.SetBusyState();
                var changeFiles = helper.GetChangelistFiles();
                if (changeFiles != null && changeFiles.Count > 0) {
                    foreach (var f in changeFiles) {
                        if (f.Action.Equals(FileAction.Delete)) {
                            var item = new FileItem();
                            item.LabelText = f.GetFileName();
                            item.ColumnContext = column;
                            item.Metadata = helper.GetFileMetaData(f.DepotPath.Path);
                            list.Add(item);
                        }
                    }
                }
            }
            return list;
        }

        public static IList<ListingItem> GetRecentFiles(ColumnDisplay column) {
            var list = Utility.GetRecentList();
            if (list != null && list.Count > 0) {
                foreach (var i in list) {
                    i.ColumnContext = column;
                }
            }
            return list;
        }

        public static ListingItem GetItem(string path, ColumnDisplay column, bool isDirectory = false) {
            ListingItem item = null;
            var helper = Utility.GetPerforceHelper();
            if (isDirectory) {
                item = new FolderItem();
                if (path.StartsWith("//")) {
                    item.DepotPath = path;
                    var paths = new List<string>();
                    paths.Add(path);
                    var mappings = helper.GetClientMappings(paths, true);
                    if (mappings.ContainsKey(path)) {
                        item.ClientPath = mappings[path];
                        item.IsMapped = true;
                    } else {
                        item.ClientPath = helper.PredictClientPath(path);
                        item.IsMapped = false;
                    }
                } else {
                    item.ClientPath = path;
                    var paths = new List<string>();
                    paths.Add(path);
                    var mappings = helper.GetDepotMappings(paths, true);
                    if (mappings.ContainsKey(path)) {
                        item.DepotPath = mappings[path];
                        item.IsMapped = true;
                    } else {
                        item.DepotPath = Constants.DUMMY_DEPOT_PATH;
                        item.IsMapped = false;
                    }
                }
                var parts = path.Split('/');
                item.LabelText = parts[parts.Length - 1];
            } else {
                var md = helper.GetFileMetaData(path);
                item = new FileItem();
                item.LabelText = md.GetFileName();
                (item as FileItem).Metadata = md;
            }
            item.ColumnContext = column;
            return item;
        }

        public static ListingItem FindItemInList(string label, IList<ListingItem> list) {
            ListingItem item = null;
            foreach (var i in list) {
                if (label.Equals(i.LabelText, StringComparison.CurrentCultureIgnoreCase)) {
                    item = i;
                    break;
                }
            }
            return item;
        }

        public static string HexToString(string hex) {
            var byteArray = HexToBytes(hex);
            return System.Text.Encoding.UTF8.GetString(byteArray);
        }

        public static byte[] HexToBytes(string value) {
            SoapHexBinary shb = SoapHexBinary.Parse(value);
            return shb.Value;
        }

        public static string BytesToHex(byte[] value) {
            SoapHexBinary shb = new SoapHexBinary(value);
            return shb.ToString();
        }

        public static byte[] GetBytesFromFile(string fullFilePath) {
            // this method is limited to 2^32 byte files (4.2 GB)
            FileStream fs = null;
            try {
                fs = System.IO.File.OpenRead(fullFilePath);
                byte[] bytes = new byte[fs.Length];
                fs.Read(bytes, 0, Convert.ToInt32(fs.Length));
                return bytes;
            } finally {
                if (fs != null) {
                    fs.Close();
                }
            }
        }
    }
}
